@@ -1,3 +1,6 @@
+import os
+import sys
+
 from .config_parser import Config
 from .config_writer import write_config
 from types import SimpleNamespace
@@ -14,6 +17,13 @@ def parse_config(config_paths: Optional[List[str]] = None) -> SimpleNamespace:
     Returns:
         SimpleNamespace: A namespace object with config and env values accessible via dot notation.
     """
+    if getattr(sys, 'frozen', False):
+        # If running as an EXE
+        run_dir = os.path.dirname(sys.executable)
+
+        # Change directory to executable root.
+        os.chdir(run_dir)
+
     if config_paths:
         Config().reload(config_paths)
 
