@@ -16,6 +16,15 @@ from ..logger import Logger, LogWrapper
 T = TypeVar("T")
 
 # Classes for storing Merge request and commit metadata
+@dataclass
+class MergeRequestActivity:
+    id: int
+    body: str
+    type: Optional[str]
+    author: str
+    created_at: str
+
+
 @dataclass
 class MergeRequestMetadata:
     parent_branch_name: str
@@ -26,6 +35,7 @@ class MergeRequestMetadata:
     merge_request_state: str
     merge_request_url: str
     branch_changed_files: Optional[List[str]] = None
+    merge_request_activity: Optional[List[MergeRequestActivity]] = None
 
 
 @dataclass
@@ -34,6 +44,7 @@ class CommitMetadata:
     commit_title: str
     commit_author_name: str
     commit_id: str
+    commit_date: str
 
 
 def handle_gitlab_errors(
@@ -116,9 +127,10 @@ class GitlabClient(UtilityBase):
         self.gl = gitlab.Gitlab(url, private_token=private_token, timeout=timeout, retry_transient_errors=True)
         self.gl.auth()
 
-    def _warn_large(self, resource, project_path: str, label: str) -> None:
+    def _warn_large(self, resource: Any, project_path: str, label: str) -> None:
         """Warn if resource count exceeds the large repo threshold."""
         count = len(resource.list(per_page=100))
+        
         if count > self.large_repo_threshold:
             self._log_warning(f"Project '{project_path}' has {count}+ {label}. all=True may be slow.")
 
@@ -141,7 +153,7 @@ class GitlabClient(UtilityBase):
         Get a GitLab project by path or ID.
 
         Args:
-            project_path: Project path or ID.
+            project_path (str): Project path or ID.
 
         Returns:
             Project object or None if not found.
@@ -154,8 +166,8 @@ class GitlabClient(UtilityBase):
         List branch names in a project.
 
         Args:
-            project_path: Project path or ID.
-            max_branches: Max branches to return. If None, return all.
+            project_path (str): Project path or ID.
+            max_branches (Optional, int): Max branches to return. If None, return all.
 
         Returns:
             List of branch names.
@@ -177,8 +189,8 @@ class GitlabClient(UtilityBase):
         Return N most recently updated branch names in the project.
 
         Args:
-            project_path: Project path or ID.
-            num_branches: Number of recent branches to return.
+            project_path (str): Project path or ID.
+            num_branches (int, default = 10): Number of recent branches to return.
 
         Returns:
             List of branch names.
@@ -202,21 +214,20 @@ class GitlabClient(UtilityBase):
         branch_name: str,
         state: str = "all",
         get_changed_files: bool = False,
+        get_activity: bool = False,  # NEW ARGUMENT
     ) -> Optional[List[MergeRequestMetadata]]:
         """
         Get all the merge requests for a branch.
 
         Args:
-            project_path: Project path or ID.
-            branch_name: Branch to check.
-            state: MR state ('opened', 'closed', 'merged', 'all').
-            get_changed_files: If True, retrieve the list of changed files per MR.
-                            If False, 'branch_changed_files' will be an empty list.
+            project_path (str): Project path or ID.
+            branch_name (str): Branch to check.
+            state (str, default = all): MR state ('opened', 'closed', 'merged', 'all').
+            get_changed_files (bool, default = False): If True, retrieve the list of changed files per MR.
+            get_activity (bool, default = False): If True, retrieve activity (notes) for the MR.
 
         Returns:
             List of MergeRequestMetadata object(s) or None.
-            If 'get_changed_files' is True and changed files cannot be retrieved,
-            'branch_changed_files' will be set to an empty list for that MR.
         """
         if not (project := self.get_project(project_path)):
             return None
@@ -239,8 +250,26 @@ class GitlabClient(UtilityBase):
                         changes = mr_full.changes()
                         branch_changed_files = [change['new_path'] for change in changes['changes']]
                     except Exception:
-                        # Could not retrieve changed files (too large, etc.)
                         branch_changed_files = []
+                else:
+                    mr_full = project.mergerequests.get(mr.iid, lazy=False)
+
+                merge_request_activity = None
+                if get_activity:
+                    try:
+                        notes = mr_full.notes.list(all=True)
+                        merge_request_activity = [
+                            MergeRequestActivity(
+                                id=note.id,
+                                body=note.body,
+                                type=getattr(note, 'type', None),
+                                author=note.author['name'],
+                                created_at=note.created_at,
+                            )
+                            for note in notes
+                        ]
+                    except Exception:
+                        merge_request_activity = []
 
                 result.append(
                     MergeRequestMetadata(
@@ -251,7 +280,8 @@ class GitlabClient(UtilityBase):
                         merge_request_description=mr.description,
                         merge_request_state=mr.state,
                         merge_request_url=mr.web_url,
-                        branch_changed_files=branch_changed_files
+                        branch_changed_files=branch_changed_files,
+                        merge_request_activity=merge_request_activity,
                     )
                 )
             return result
@@ -264,8 +294,8 @@ class GitlabClient(UtilityBase):
         List tag names in a project.
 
         Args:
-            project_path: Project path or ID.
-            max_tags: Max tags to return. If None, return all.
+            project_path (str): Project path or ID.
+            max_tags (Optional, int): Max tags to return. If None, return all.
 
         Returns:
             List of tag names.
@@ -289,9 +319,9 @@ class GitlabClient(UtilityBase):
         List all files changed between two tags.
 
         Args:
-            project_path: Project path or ID.
-            from_tag: The base tag.
-            to_tag: The target tag.
+            project_path (str): Project path or ID.
+            from_tag (str): The base tag.
+            to_tag (str): The target tag.
 
         Returns:
             List of changed file paths.
@@ -304,6 +334,39 @@ class GitlabClient(UtilityBase):
         changed_files = {diff['new_path'] for diff in diffs}
         return list(changed_files)
 
+    @handle_gitlab_errors(default_return=None)
+    def get_commit_metadata(
+        self, 
+        project_path: str, 
+        commit_hash: str
+    ) -> Optional[CommitMetadata]:
+        """
+        Get metadata for a single commit by hash.
+
+        Args:
+            project_path (str): Project path or ID.
+            commit_hash (str): Commit hash or short hash.
+
+        Returns:
+            CommitMetadata or None.
+        """
+        if not (project := self.get_project(project_path)):
+            return None
+
+        try:
+            commit = project.commits.get(commit_hash)
+        except Exception:
+            self._log_exception(f"Error occurred while requesting commit metadata for `{commit_hash}`")
+            return None
+        
+        return CommitMetadata(
+            commit_short_id=commit.short_id,
+            commit_title=commit.title,
+            commit_author_name=commit.author_name,
+            commit_id=commit.id,
+            commit_date=commit.committed_date
+        )
+
     @handle_gitlab_errors(default_return=[])
     def list_commits(
         self, project_path: str, branch: str, max_commits: Optional[int] = None
@@ -312,9 +375,9 @@ class GitlabClient(UtilityBase):
         List commits on a branch.
 
         Args:
-            project_path: Project path or ID.
-            branch: Branch to list commits from.
-            max_commits: Max commits to return. If None, return all.
+            project_path (str): Project path or ID.
+            branch (str): Branch to list commits from.
+            max_commits (Optional, int): Max commits to return. If None, return all.
 
         Returns:
             List of CommitMetadata.
@@ -333,7 +396,8 @@ class GitlabClient(UtilityBase):
                 commit_short_id=commit.short_id,
                 commit_title=commit.title,
                 commit_author_name=commit.author_name,
-                commit_id=commit.id
+                commit_id=commit.id,
+                commit_date=commit.committed_date
             )
             for commit in commits
         ]
@@ -346,9 +410,9 @@ class GitlabClient(UtilityBase):
         List commits unique to compare_branch relative to base_branch.
 
         Args:
-            project_path: Project path or ID.
-            base_branch: Base branch.
-            compare_branch: Branch to find unique commits in.
+            project_path (str): Project path or ID.
+            base_branch (str): Base branch.
+            compare_branch (str): Branch to find unique commits in.
 
         Returns:
             List of CommitMetadata.
@@ -358,6 +422,7 @@ class GitlabClient(UtilityBase):
         
         comparison = project.repository_compare(from_=base_branch, to=compare_branch)
         commits = comparison['commits']
+
         return [
             CommitMetadata(
                 commit_short_id=commit['short_id'],
@@ -374,8 +439,8 @@ class GitlabClient(UtilityBase):
         Get list of files changed in a commit.
 
         Args:
-            project_path: Project path or ID.
-            commit_id: Commit SHA or ID.
+            project_path (str): Project path or ID.
+            commit_id (str): Commit SHA or ID.
 
         Returns:
             List of changed file paths.
@@ -394,9 +459,9 @@ class GitlabClient(UtilityBase):
         Clone a remote repository to local path (or open if already exists).
 
         Args:
-            project_url: Remote repository path (after the GitLab base URL).
-            local_path: Local directory for the clone.
-            branch: Optional branch to check out.
+            project_url (str): Remote repository path (after the GitLab base URL).
+            local_path (str): Local directory for the clone.
+            branch (Optional, str): Optional branch to check out.
 
         Returns:
             GitPython Repo object or None if failure.
@@ -419,8 +484,8 @@ class GitlabClient(UtilityBase):
         Check out a branch in a local repository.
 
         Args:
-            local_path: Local repository path.
-            branch: Branch name to check out.
+            local_path (str): Local repository path.
+            branch (str): Branch name to check out.
         """
         repo = Repo(local_path)
 
@@ -436,9 +501,9 @@ class GitlabClient(UtilityBase):
         Check out a tag in a local repository. Optionally create and switch to a new branch at the tag.
 
         Args:
-            local_path: Local repository path.
-            tag: Tag name to check out.
-            new_branch: If provided, create this branch from the tag and check it out.
+            local_path (str): Local repository path.
+            tag (str): Tag name to check out.
+            new_branch (Optional, str): If provided, create this branch from the tag and check it out.
         """
         repo = Repo(local_path)
 
@@ -460,10 +525,10 @@ class GitlabClient(UtilityBase):
         Download and extract artifacts from specified jobs in a pipeline.
 
         Args:
-            project_path: Project path or ID.
-            pipeline_id: Pipeline ID.
-            jobs_of_interest: List of job names whose artifacts to download.
-            output_folder: Directory to save artifacts.
+            project_path (str): Project path or ID.
+            pipeline_id (int): Pipeline ID.
+            jobs_of_interest (list[str]): List of job names whose artifacts to download.
+            output_folder (str): Directory to save artifacts.
         """
         import shutil
 
@@ -487,13 +552,19 @@ class GitlabClient(UtilityBase):
                 os.makedirs(job_dir, exist_ok=True)
                 zip_path = os.path.join(job_dir, "artifacts.zip")
                 try:
+
+                    # Get the full job object
+                    full_job = project.jobs.get(job.id)
+
                     # Download artifact
                     with open(zip_path, "wb") as f:
-                        job.artifacts(streamed=True, action=f.write)
+                        f.write(full_job.artifacts())
+
                     # Unpack artifact
                     with zipfile.ZipFile(zip_path, 'r') as z:
                         z.extractall(path=job_dir)
                     os.remove(zip_path)
+
                     self._log_info(f"Downloaded and extracted artifacts for job: {job.name}")
                 except Exception as e:
                     self._log_error(f"Failed to process artifacts for job {job.name}: {e}")
